from . import panorama, restapi, xmlapi
from .panorama import Panorama
from .restapi.restapi import PanoramaClient
from .xmlapi import XMLApi, types

__all__ = [
    "panorama",
    "restapi",
    "xmlapi",
]



from nagra_panorama_api.utils import (
    first,
)

from .address import Address
from .device import Device, HAInfo, VPNFlow
from .interface import (
    AggregateEthernet,
    Ethernet,
    GenericInterface,
    Interface,
    Layer2,
    Layer3,
    Loopback,
    Vlan,
)
from .job import Job, JobResult
from .software import SoftwareVersion

# from nagra_panorama_api.xmlapi.types.utils import (
#     Datetime,
#     mksx,
#     parse_datetime,
#     parse_time,
#     pd,
# )
# from nagra_panorama_api.xmlapi.utils import (
#     el2dict,
#     pprint,
# )



import logging
from datetime import datetime
from pathlib import Path

import yaml

DEFAULT_OUTPUT_DIRECTOR = "upgrade_logs"


def ensure_output_directory(output, group):
    if output == "":
        return output
    if output is None:
        output = DEFAULT_OUTPUT_DIRECTOR

    # Check that the destination directory is not a file
    logs_output = Path(output).resolve()
    if logs_output.is_file():
        logging.error(f"{logs_output} is a file, it must be a folder or not exists")
        exit(1)

    now = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    group_folder = f"{group}_{now}"
    # Check that the group directory is not a file
    group_logs_output = logs_output / Path(group_folder)
    if group_logs_output.is_file():
        logging.error(
            f"""\
Destination '{logs_output}' already contains a file named '{group_folder}'.
Please rename/delete/move file '{group_logs_output}' or change the destination folder
"""
        )
        exit(1)
    group_logs_output.mkdir(parents=True)
    return group_logs_output


def ensure_group_load(file, group):
    file = Path(file).resolve()
    if not file.exists():
        logging.error(f"File {file} does not exist")
        exit(1)
    if file.suffix not in (".yml", ".yaml"):
        logging.error("File must be in yaml format")
        exit(1)
    try:
        data = yaml.safe_load(file.read_text())
        groups = data["groups"]
        group_data = groups.get(group)
        if group_data:
            return group_data
        logging.error(f"Group '{group}' does not exist ini file {file}")
        defined_groups = groups.keys()
        if defined_groups:
            defined_groups = ", ".join(sorted(defined_groups))
            logging.info(f"Possible groups are: {defined_groups}")
        exit(1)
    except Exception as e:
        logging.error(
            f"""\
An error occured when parsing the file '{file}' and/or retrieving the data of group '{group}'
Make sure that the file is formatted correctly and that the group exists
{e}
"""
        )
        exit(1)



import os
import re

URL_REG = re.compile("(http://|https://)?([^:/]*)(?::(\d+))?(/.*)?")


def first(iterable, default=None):
    return next(iter(iterable), default)


def clean_url_host(url):
    """
    Clean the url:
    * Add scheme if missing
    * Remove trailing slash and path
    * Extract the host
    """
    scheme, host, port, _ = URL_REG.match(url).groups()
    scheme = scheme or "https://"
    url = scheme + host + (f":{port}" if port else "")
    return url, host, port


# =============================================
def getenv(*varnames, raise_exception=True, default=None):
    for var in varnames:
        val = os.environ.get(var)
        if val is not None:
            return val.strip()
    if raise_exception:
        raise Exception(
            f"None of the following environment variables are defined: {', '.join(varnames)}"
        )
    return default


# =======================================================================================================


def get_credentials_from_env(raise_exception=False):
    host = getenv("PANORAMA_HOST", raise_exception=raise_exception)
    apikey = getenv("PANORAMA_APIKEY", "PANO_APIKEY", raise_exception=raise_exception)
    return host, apikey



"""
This example will download and install the latest version of the software.

Python requirements:
- ping3
"""

import logging
import os
import re
from pathlib import Path

# from nagra_network_misc_utils.logger import set_default_logger
from rich.logging import RichHandler

# set_default_logger()
logging.getLogger().setLevel(logging.INFO)
# logging.getLogger().setLevel(logging.DEBUG)

SLUGIFY_REG = re.compile("\w+")


def slugify(name, join="_"):
    return join.join(SLUGIFY_REG.findall(name.strip().lower()))


def getenv(*varnames, exit_on_failure=True, default=None):
    for var in varnames:
        val = os.environ.get(var)
        if val is not None:
            return val.strip()
    logging.error(
        f"None of the following environment variables are defined: {', '.join(varnames)}"
    )
    if exit_on_failure:
        exit(1)
    return default


# %(asctime)s - %(name)s - %(levelname)s - %(message)s
DEFAULT_LOGGER_FORMAT = "[%(levelname)s] %(message)s"

LOGGER_REGISTRY = {}


def get_logger(name, file=None, level=logging.INFO):
    global LOGGER_REGISTRY
    logger = LOGGER_REGISTRY.get(name)
    if logger is not None:
        return logger
    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.propagate = False  # Prevent calling the root logger
    logger.handlers.clear()
    formatter = logging.Formatter(DEFAULT_LOGGER_FORMAT)
    if file is not None:
        file = Path(file).resolve()
        file.parent.mkdir(parents=True, exist_ok=True)
        file_handler = logging.FileHandler(file, mode="w")
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    stream_handler = RichHandler()
    logger.addHandler(stream_handler)
    LOGGER_REGISTRY[name] = logger
    return logger



from ipaddress import IPv4Network, IPv6Network
from typing import List, Union

from pydantic import AliasChoices, BaseModel, ConfigDict, Field

IPNetwork = Union[IPv4Network, IPv6Network]


class Address(BaseModel):
    model_config = ConfigDict(extra="allow")

    device_group: str
    name: str = Field(validation_alias=AliasChoices("@name", "name"))
    prefix: str
    ip_netmask: str
    ip_network: IPNetwork
    ip_range: str
    fqdn: str
    tags: List[str]

    # name: str = Field(validation_alias=AliasChoices("@name", "name"))
    # device_group: str = Field(alias="device-group")
    # rulebase: Literal["pre-rulebase", "post-rulebase"] = Field(alias="rulebase")
    # ttype: str = Field(alias="type")
    # state: Optional[str] = Field(alias="rule-state")
    # modification: datetime = Field(alias="rule-modification-timestamp")
    # creation: datetime = Field(alias="rule-creation-timestamp")
    # all_connected: bool = Field(alias="all-connected")

    # @field_validator("modification", "creation", mode="before")
    # @classmethod
    # def ensure_datetime(cls, v: Any):
    #     if not isinstance(v, int):
    #         v = int(v)
    #     return datetime.fromtimestamp(v)



import logging
from itertools import chain, product
from multiprocessing.pool import ThreadPool as Pool
from typing import List, Optional, Tuple, Union

from nagra_panorama_api.utils import clean_url_host, first, get_credentials_from_env

from . import types
from .base import _get_rule_use_cmd, get_tree, raw_request
from .utils import (
    Element,
    diff_patch,
    el2dict,
    extend_element,
    map_dicts,
    wait,
)


class XMLApi:
    def __init__(
        self,
        host=None,
        api_key=None,
        ispanorama=None,
        verify=False,
        timeout=None,
        logger=None,
    ):
        env_host, env_apikey = get_credentials_from_env()
        host = host or env_host
        api_key = api_key or env_apikey
        if not host:
            raise Exception("Missing Host")
        if not api_key:
            raise Exception("Missing API Key")
        host, _, _ = clean_url_host(host)

        self._host = host
        self._api_key = api_key
        self._url = f"{host}/api"
        self._verify = verify
        self._timeout = timeout
        self._ispanorama = ispanorama
        self.logger = logger or logging

    def _request(
        self,
        type,
        method="GET",
        vsys=None,
        params=None,
        remove_blank_text=True,
        verify=None,
        parse=True,
        stream=None,
        timeout=None,
    ):
        if verify is None:
            verify = self._verify
        if timeout is None:
            timeout = self._timeout
        headers = {"X-PAN-KEY": self._api_key}
        return raw_request(
            self._url,
            type,
            method,
            vsys=vsys,
            params=params,
            headers=headers,
            remove_blank_text=remove_blank_text,
            verify=verify,
            logger=self.logger,
            parse=parse,
            stream=stream,
            timeout=timeout,
        )

    # https://docs.paloaltonetworks.com/pan-os/9-1/pan-os-panorama-api/pan-os-xml-api-request-types/export-files-api
    # https://knowledgebase.paloaltonetworks.com/KCSArticleDetail?id=kA10g000000ClaOCAS#:~:text=From%20the%20GUI%2C%20go%20to%20Device%20%3E%20Setup,%3E%20scp%20export%20configuration%20%5Btab%20for%20command%20help%5D
    def _export_request(
        self,
        category,
        method="GET",
        params=None,
        verify=None,
        stream=None,
        timeout=None,
    ):
        if params is None:
            params = {}
        params = {"category": category, **params}
        return self._request(
            "export",
            method=method,
            params=params,
            verify=verify,
            parse=False,
            stream=stream,
            timeout=timeout,
        ).content

    def export_configuration(
        self,
        verify=None,
        timeout=None,
    ) -> Element:
        return self._export_request(
            category="configuration",
            verify=verify,
            timeout=timeout,
        )

    def export_device_state(
        self,
        verify=None,
        timeout=None,
    ) -> Element:
        return self._export_request(
            category="device-state",
            verify=verify,
            timeout=timeout,
        )

    def _conf_request(
        self,
        xpath,
        action="get",
        method="GET",
        vsys=None,
        params=None,
        remove_blank_text=True,
        verify=None,
        timeout=None,
    ) -> Element:
        if params is None:
            params = {}
        params = {"action": action, "xpath": xpath, **params}
        return self._request(
            "config",
            method=method,
            vsys=vsys,
            params=params,
            remove_blank_text=remove_blank_text,
            verify=verify,
            timeout=timeout,
        )

    def _op_request(
        self,
        cmd,
        method="POST",
        vsys=None,
        params=None,
        remove_blank_text=True,
        verify=None,
        timeout=None,
    ) -> Element:
        if params is None:
            params = {}
        params = {"cmd": cmd, **params}
        return self._request(
            "op",
            method=method,
            vsys=vsys,
            params=params,
            remove_blank_text=remove_blank_text,
            verify=verify,
            timeout=timeout,
        )

    def _commit_request(
        self,
        cmd,
        method="POST",
        params=None,
        remove_blank_text=True,
        verify=None,
        timeout=None,
    ):
        if params is None:
            params = {}
        params = {"cmd": cmd, **params}
        return self._request(
            "commit",
            method=method,
            params=params,
            remove_blank_text=remove_blank_text,
            verify=verify,
            timeout=timeout,
        )

    # https://docs.paloaltonetworks.com/pan-os/9-1/pan-os-panorama-api/get-started-with-the-pan-os-xml-api/get-your-api-key
    def generate_apikey(self, username, password: str) -> str:
        """
        Generate a new API-Key for the user connected.
        """
        params = {"user": username, "password": password}
        return self._request(
            "keygen",
            method="POST",
            params=params,
        ).xpath(".//key/text()")[0]

    # https://docs.paloaltonetworks.com/pan-os/9-1/pan-os-panorama-api/pan-os-xml-api-request-types/get-version-info-api
    def api_version(self):
        return el2dict(
            self._request(
                "version",
                method="POST",
            ).xpath(".//result")[0]
        )["result"]

    def configuration(
        self,
        xpath,
        action="get",
        method="GET",
        params=None,
        remove_blank_text=True,
    ):
        return self._conf_request(
            xpath,
            action=action,
            method=method,
            params=params,
            remove_blank_text=remove_blank_text,
        )

    def operation(
        self,
        cmd,
        method="POST",
        params=None,
        remove_blank_text=True,
    ):
        return self._op_request(
            cmd,
            method=method,
            params=params,
            remove_blank_text=remove_blank_text,
        )

    def _check_is_panorama(self) -> bool:
        try:
            self.configuration("/config/panorama/vsys")
            return False
        except Exception:
            return True

    @property
    def ispanorama(self):
        if self._ispanorama is None:
            self._ispanorama = self._check_is_panorama()
        return self._ispanorama

    def get_tree(self, extended=False) -> Element:
        """
        Return the running configuration
        The differences with `running_config` are not known
        """
        tree = get_tree(
            self._host, self._api_key, verify=self._verify, logger=self.logger
        )
        if extended:
            self._extend_tree_information(tree)
        return tree

    def _get_rule_use(self, device_group, position, rule_type, number: int = 200):
        results = []
        for i in range(100):
            cmd = _get_rule_use_cmd(
                device_group,
                position,
                rule_type,
                i * number,
                number,
            )
            res = self._op_request(cmd).xpath("result")[0]
            total_count = int(res.attrib["total-count"])
            results.extend(res.xpath("entry"))
            if len(results) >= total_count:
                break
        return results

    def get_rule_use(self, tree=None, max_threads: Optional[int] = None):
        if tree is None:
            tree = self.get_tree()
        device_groups = tree.xpath("devices/*/device-group/*/@name")
        positions = ("pre", "post")
        # rule_types = tuple({x.tag for x in tree.xpath(
        # "devices/*/device-group/*"
        # "/*[self::post-rulebase or self::pre-rulebase]/*")})
        rule_types = ("security", "pbf", "nat", "application-override")
        args_list = list(product(device_groups, positions, rule_types))

        def func(args):
            return self._get_rule_use(*args)

        threads = len(args_list)
        threads = min(max_threads or threads, threads)
        with Pool(len(args_list)) as pool:
            data = pool.map(func, args_list)
        return [entry for entry_list in data for entry in entry_list]

    def _get_rule_hit_count(self, device_group, rulebase, rule_type):
        cmd = (
            "<show><rule-hit-count><device-group>"
            f"<entry name='{device_group}'><{rulebase}><entry name='{rule_type}'>"
            f"<rules><all/></rules></entry></{rulebase}></entry>"
            "</device-group></rule-hit-count></show>"
        )
        res = self._op_request(cmd)
        entries = res.xpath(".//rules/entry") or []
        # return entries
        return [(device_group, rulebase, rule_type, e) for e in entries]

    def get_rule_hit_count(self, tree=None, max_threads=None):
        if tree is None:
            tree = self.get_tree()
        device_groups = tree.xpath("devices/*/device-group/*/@name")
        rulebases = ("pre-rulebase", "post-rulebase")
        rule_types = ("security", "pbf", "nat", "application-override")
        args_list = list(product(device_groups, rulebases, rule_types))

        def func(args):
            return self._get_rule_hit_count(*args)

        threads = len(args_list)
        threads = min(max_threads or threads, threads)
        with Pool(len(args_list)) as pool:
            data = pool.map(func, args_list)
        return [entry for entry_list in data for entry in entry_list]

    def _extend_tree_information(
        self,
        tree,
        extended=None,
        max_threads=None,
    ):
        """
        Incorporate usage statistics into the configuration.
        tree: the configuration as a XML object
        extended: rule-use data (if not provided, the function will retrieve them automatically)
        """
        if extended is None:
            extended = self.get_rule_use(tree, max_threads=max_threads)
        rules = tree.xpath(
            ".//device-group/entry/"
            "*[self::pre-rulebase or self::post-rulebase]/*/rules/entry[@uuid]",
        )
        ext_dict = {x.attrib.get("uuid"): x for x in extended}
        rules_dict = {x.attrib["uuid"]: x for x in rules}
        for ext, rule in map_dicts(ext_dict, rules_dict):
            extend_element(rule, ext)
            # NOTE: Do not use rule.extend(ext)
            # => This is causing duplicates entries
        return tree, extended

    def get(self, xpath: str):
        """
        This will retrieve the xml definition based on the xpath
        The xpath doesn't need to be exact
        and can select multiple values at once.
        Still, it must at least speciy /config at is begining
        """
        return self._conf_request(xpath, action="show", method="GET")

    def delete(self, xpath: str):
        """
        This will REMOVE the xml definition at the provided xpath.
        The xpath must be exact.
        """
        return self._conf_request(
            xpath,
            action="delete",
            method="DELETE",
        )

    def create(self, xpath: str, xml_definition):
        """
        This will ADD the xml definition
        INSIDE the element at the provided xpath.
        The xpath must be exact.
        """
        # https://docs.paloaltonetworks.com/pan-os/9-1/pan-os-panorama-api/pan-os-xml-api-request-types/configuration-api/set-configuration
        params = {"element": xml_definition}
        return self._conf_request(
            xpath,
            action="set",
            method="POST",
            params=params,
        )

    def update(self, xpath: str, xml_definition):
        """
        This will REPLACE the xml definition
        INSTEAD of the element at the provided xpath
        The xpath must be exact.
        Nb: We can pull the whole config, update it locally,
        and push the final result
        """
        # https://docs.paloaltonetworks.com/pan-os/9-1/pan-os-panorama-api/pan-os-xml-api-request-types/configuration-api/set-configuration
        params = {"element": xml_definition}
        return self._conf_request(
            xpath,
            action="edit",
            method="POST",
            params=params,
        )

    def revert_changes(self, skip_validated: bool = False):
        """
        Revert all the changes made on Panorama.
        NOTE:
        - This only applies on non-commited changes.
        - This revert everything (not scoped by users)

        skip_validated: Do not revert changes that were validated
        """
        skip = "<skip-validate>yes</skip-validate>" if skip_validated else ""
        cmd = f"<revert><config>{skip}</config></revert>"
        return self._op_request(cmd)

    def validate_changes(self):
        """
        Validated all the changes currently made
        """
        cmd = "<validate><full></full></validate>"
        return self._op_request(cmd)

    def _raw_get_push_scope(self, admin=None):
        """
        Gives detailed information about pending changes
        (e.g. xpath, owner, action, ...)
        """
        filter = f"<admin><member>{admin}</member></admin>" if admin else ""
        cmd = f"<show><config><push-scope>{filter}</push-scope></config></show>"
        return self._op_request(cmd)

    def get_push_scope_devicegroups(self, admin=None):
        """
        Gives detailed information about pending changes
        (e.g. xpath, owner, action, ...)
        """
        scope = self._raw_get_push_scope(admin=admin)
        return list(set(scope.xpath(".//objects/entry[@loc-type='device-group']/@loc")))

    def uncommited_changes(self):
        """
        Gives detailed information about pending changes
        (e.g. xpath, owner, action, ...)
        """
        cmd = "<show><config><list><changes></changes></list></config></show>"
        return self._op_request(cmd)

    def uncommited_changes_patch(self):
        """
        Gives detailed information about pending changes
        (e.g. xpath, owner, action, ...)
        """
        candidate = self.candidate_config()
        running = self.running_config()
        return diff_patch(running, candidate)

    def uncommited_changes_summary(self, admin=None):
        """
        Only gives the concern device groups
        """
        admin = (
            f"<partial><admin><member>{admin}</member></admin></partial>"
            if admin
            else ""
        )
        cmd = f"<show><config><list><change-summary>{admin}</change-summary></list></config></show>"
        return self._op_request(cmd)

    def pending_changes(self):
        """
        Result content is either 'yes' or 'no'
        """
        cmd = "<check><pending-changes></pending-changes></check>"
        return self._op_request(cmd)

    def save_config(self, name):
        """
        Create a named snapshot of the current configuration
        """
        cmd = f"<save><config><to>{name}</to></config></save>"
        return "\n".join(self._op_request(cmd).xpath(".//result/text()"))

    def save_device_state(self):
        """
        Create a snapshot of the current device state
        """
        cmd = "<save><device-state></device-state></save>"
        return "\n".join(self._op_request(cmd).xpath(".//result/text()"))

    def get_named_configuration(self, name):
        """
        Get the configuration from a named snapshot as an XML object
        """
        cmd = f"<show><config><saved>{name}</saved></config></show>"
        return self._op_request(cmd, remove_blank_text=False).xpath("./result/config")[
            0
        ]

    def candidate_config(self) -> Element:
        """
        Get the configuration to be commited as an XML object
        """
        cmd = "<show><config><candidate></candidate></config></show>"
        return self._op_request(cmd, remove_blank_text=False)

    def running_config(self) -> Element:
        """
        Get the current running configuration as an XML object
        """
        cmd = "<show><config><running></running></config></show>"
        return self._op_request(cmd, remove_blank_text=False)

    def _raw_get_jobs(self, job_ids: Union[None, str, List[str]] = None) -> Element:
        """
        Get information of job(s) as an XML object.
        Retrieve all jobs by default.

        If job_id is provided, then only retrieve the job requested.
        """
        filter = "<all></all>"
        if job_ids:
            if isinstance(job_ids, str):
                job_ids = [job_ids]
            filter = "".join(f"<id>{j}</id>" for j in job_ids)
        cmd = f"<show><jobs>{filter}</jobs></show>"
        return self._op_request(cmd)

    def get_jobs(self, job_ids: Union[None, str, List[str]] = None) -> List[types.Job]:
        """
        Get information of job(s)
        Retrieve all jobs by default.

        If job_id is provided, then only retrieve the job requested.
        """
        job_xmls = self._raw_get_jobs(job_ids).xpath(".//job")
        transformed = (types.Job.from_xml(x) for x in job_xmls)
        return [j for j in transformed if j]

    def get_job(self, job_id) -> types.Job:
        """
        Get information of job(s)
        Retrieve all jobs by default.

        If job_id is provided, then only retrieve the job requested.
        """
        return self.get_jobs(job_id)[0]

    def _raw_get_versions(self) -> Element:
        """
        Get the versions informations as a XML object.
        """
        cmd = "<request><system><software><check></check></software></system></request>"
        return self.operation(cmd)

    def get_versions(self) -> List[types.SoftwareVersion]:
        """
        Get the versions informations
        """
        res = self._raw_get_versions()
        return [
            types.SoftwareVersion.from_xml(entry)
            for entry in res.xpath(".//sw-updates/versions/entry")
        ]

    def wait_job_completion(self, job_id: str, waiter=None) -> types.Job:
        """
        Block until the job complete.

        job_id: the job to wait upon
        waiter: a generator that yield when a new query must be done.
                see `wait` function (the default waiter) for an example
        """
        if not waiter:
            waiter = wait()
        for _ in waiter:
            job = self.get_job(job_id)
            if job.progress >= 100:
                return job
            self.logger.info(f"Job {job_id} progress: {job.progress}")
        raise Exception("Timeout while waiting for job completion")

    def raw_get_pending_jobs(self):
        """
        Get all the jobs that are pending as a XML object
        """
        cmd = "<show><jobs><pending></pending></jobs></show>"
        return self._op_request(cmd)

    def commit_changes(self, force: bool = False):
        """
        Commit all changes
        """
        cmd = "<commit>{}</commit>".format("<force></force>" if force else "")
        return self._commit_request(cmd)

    def _lock_cmd(self, cmd, vsys, no_exception=False) -> bool:
        """
        Utility function for commands that tries to manipulate the lock
        on Panorama.
        """
        try:
            result = "".join(self._op_request(cmd, vsys=vsys).itertext())
            self.logger.debug(result)
        except Exception as e:
            if no_exception:
                self.logger.error(e)
                return False
            raise
        return True

    # https://github.com/PaloAltoNetworks/pan-os-python/blob/a6b018e3864ff313fed36c3804394e2c92ca87b3/panos/base.py#L4459
    def add_config_lock(self, comment=None, vsys="shared", no_exception=False) -> bool:
        comment = f"<comment>{comment}</comment>" if comment else ""
        cmd = f"<request><config-lock><add>{comment}</add></config-lock></request>"
        return self._lock_cmd(cmd, vsys=vsys, no_exception=no_exception)

    def remove_config_lock(self, vsys="shared", no_exception=False) -> bool:
        cmd = "<request><config-lock><remove></remove></config-lock></request>"
        return self._lock_cmd(cmd, vsys=vsys, no_exception=no_exception)

    def add_commit_lock(self, comment=None, vsys="shared", no_exception=False) -> bool:
        comment = f"<comment>{comment}</comment>" if comment else ""
        cmd = f"<request><commit-lock><add>{comment}</add></commit-lock></request>"
        return self._lock_cmd(cmd, vsys=vsys, no_exception=no_exception)

    def remove_commit_lock(self, vsys="shared", no_exception=False) -> bool:
        cmd = "<request><commit-lock><remove></remove></commit-lock></request>"
        return self._lock_cmd(cmd, vsys=vsys, no_exception=no_exception)

    def set_ha_status(self, active: bool = True, target: Optional[str] = None):
        """
        Activate or Deactivate (suspend) the HA pair.

        """
        status = "<functional></functional>" if active else "<suspend></suspend>"
        cmd = f"<request><high-availability><state>{status}</state></high-availability></request>"
        params = {"target": target} if target else None
        return self._op_request(cmd, params=params).xpath(".//result/text()")[0]

    def set_ha_preemption(self, active=True, target=None):
        """
        NOT WORKING:
        There is currently no way to deactivate the preemption using the API.
        """
        raise Exception("set_ha_preemption not implementend")

    def _raw_get_ha_info(self, state_only=False, target=None) -> Element:
        """
        Get the current state of a HA pair as a XML object.
        """
        filter = "<state></state>" if state_only else "<all></all>"
        cmd = f"<show><high-availability>{filter}</high-availability></show>"
        params = {"target": target} if target else None
        return self._op_request(cmd, params=params)

    def get_ha_info(self, state_only=False, target=None) -> types.HAInfo:
        """
        Get the current state of a HA pair as a python object.
        """
        res = self._raw_get_ha_info(state_only=state_only, target=target)
        hainfo_xml = res.xpath(".//result")[0]
        # pprint(hainfo_xml)
        return types.HAInfo.from_xml(hainfo_xml)

    def get_ha_pairs(
        self, connected=True
    ) -> Tuple[List[Tuple[types.Device, Optional[types.Device]]], List[types.Device]]:
        """
        Retrieve a tuple containing 2 values:
        1. The list of HA pairs and their members
        2. A list of devices that are not part of a HA pair
        """
        # Get all devices and index them using their serial number
        devices: List[types.Device] = self.get_devices(connected=connected)
        device_map = {d.serial: d for d in devices}

        # Create the 2 lists by iterating over the devices
        done = set()
        ha_pairs = []
        without_ha = []
        for d in devices:
            # Do not manage twice the same device
            if d.serial in done:
                continue
            # The device does not have an HA peer
            if not d.ha_peer_serial:
                without_ha.append(d)
                done.add(d.serial)
                continue
            # Get the current device's HA peer
            # push them in the ha_pairs list
            # and mark both of them as done
            peer = device_map.get(d.ha_peer_serial)
            ha_pairs.append((d, peer))
            done.update((d.serial, d.ha_peer_serial))
        return ha_pairs, without_ha

    def get_ha_pairs_map(self, connected=True):
        """
        Same as `get_ha_pairs`, but the ha_pairs are return as a map.
        This provides an easier and more readable lookup to find a pair:

        mapping, _ = client.get_ha_pairs_map()
        serial = "12345"
        pair_of_serial = mapping[serial]
        """
        ha_pairs, without_ha = self.get_ha_pairs(connected=connected)
        map = {}
        for pair in ha_pairs:
            for device in pair:
                map[device.serial] = pair
        return map, without_ha

    def get_panorama_status(self):
        """
        Get the current status of Panorama server.
        """
        cmd = "<show><panorama-status></panorama-status></show>"
        return self.operation(cmd).xpath(".//result")

    def raw_get_local_panorama(self):
        return self.configuration(
            "/config/devices/entry/deviceconfig/system/panorama/local-panorama/panorama-server"
        )

    def get_local_panorama_ip(self) -> Optional[str]:
        res = self.raw_get_local_panorama()
        return first(res.xpath("//panorama-server/text()"))

    def _raw_get_devices(self, connected=False):
        """
        Return the list of device known from Panorama as a XML object.
        NOTE: This only works if the client is a Panorama server.

        connected: only returns the devices that are connected
        """
        # This only works on Panorama, not the FW
        filter = "<connected></connected>" if connected else "<all></all>"
        cmd = f"<show><devices>{filter}</devices></show>"
        return self.operation(cmd)

    def get_devices(self, connected=False) -> List[types.Device]:
        """
        Return the list of device known from Panorama as a python structure.
        NOTE: This only works if the client is a Panorama server.

        connected: only returns the devices that are connected
        """
        res = self._raw_get_devices(connected=connected)
        entries = res.xpath(".//devices/entry")
        devices = (types.Device.from_xml(e) for e in entries)
        return [d for d in devices if d]

    def _raw_get_dg_hierarchy(self):
        """
        Return the hierarchy of device groups as a XML object.
        """
        cmd = "<show><dg-hierarchy></dg-hierarchy></show>"
        return self.operation(cmd)

    def get_plan_dg_hierarchy(self, recursive=False):
        """
        Return the hierarchy of device groups as a dict.
        The keys are the names of the device groups.

        The values are the children device groups and depends on the recursive parameter.
        recursive: if False, the values are only the direct children of the device group.
            Otherwise, the values are all the descendant device groups.
        """
        devicegroups = {}  # name: children
        hierarchy = self._raw_get_dg_hierarchy().xpath(".//dg-hierarchy")[0]
        xpath = ".//dg" if recursive else "./dg"
        for dg in hierarchy.xpath(".//dg"):
            devicegroups[dg.attrib["name"]] = [
                x.attrib["name"] for x in dg.xpath(xpath)
            ]
        return devicegroups

    def _raw_get_devicegroups(self):
        """
        Return the list of device groups as a XML object.
        """
        cmd = "<show><devicegroups></devicegroups></show>"
        return self.operation(cmd)

    def get_devicegroups_name(
        self,
        parents=None,
        with_connected_devices=None,
    ):
        """
        This returns the names of the devicegroups:
        - parents: the returned list will only contain children of the provided parents (parents included)
        - with_devices: the returned list will only contain devicegroups that have direct devices under them
        """
        devicegroups = self._raw_get_devicegroups().xpath(".//devicegroups/entry")
        if with_connected_devices:
            names = [
                dg.attrib["name"]
                for dg in devicegroups
                if dg.xpath("./devices/entry/connected[text() = 'yes']")
            ]
        else:
            names = [dg.attrib["name"] for dg in devicegroups]
        if parents:
            hierarchy = self.get_plan_dg_hierarchy(recursive=True)
            tokeep = set(chain(*(hierarchy.get(p, []) for p in parents))) | set(parents)
            names = list(set(names) & tokeep)
        return names

    def _raw_get_addresses(self):
        """
        Return the list of addresses known from Panorama as a XML object.
        NOTE: This only works if the client is a Firewall.
        """
        if self.ispanorama:
            return self.configuration(
                "/config/devices/entry/device-group/entry/address"
            )
        return self.configuration("/config/panorama/vsys//address")

    def get_addresses(self) -> List[types.Address]:
        """
        Return the list of addresses known from Panorama as a python structure.
        NOTE: This only works if the client is a Firewall.
        """
        res = self._raw_get_addresses()
        addresses = res.xpath(".//address/entry")
        return [types.Address.from_xml(i) for i in addresses]

    def _raw_get_interfaces(self):
        """
        Return the list of interfaces known from Panorama as a XML object.
        NOTE: This only works if the client is a Firewall.
        """
        return self.configuration("/config/devices/entry/network/interface")

    def get_interfaces(self) -> List[types.Interface]:
        """
        Return the list of interface known from Panorama as a python structure.
        NOTE: This only works if the client is a Firewall.
        """
        res = self._raw_get_interfaces()
        interfaces = res.xpath(".//interface")
        return [types.Interface.from_xml(i) for i in interfaces]

    def _raw_get_templates(self, name=None):
        """
        Return the synchronization status the templates per devices as a XML object.
        A device is in sync if it is up-to-date with the current version on Panorama.
        NOTE: This only works on Panorama.
        """
        # This only works on Panorama, not the FW
        filter = f"<name>{name}</name>" if name else ""
        cmd = f"<show><templates>{filter}</templates></show>"
        return self.operation(cmd)

    def get_templates_sync_status(self):
        """
        Return the synchronization status the templates per devices
        A device is in sync if it is up-to-date with the current version on Panorama.
        NOTE: This only works on Panorama.

        The result is a list of tuple of 3 values:
        1. the template's name
        2. the device's name
        3. the sync status
        """
        res = self._raw_get_templates()
        statuses = []
        for entry in res.xpath("./result/templates/entry"):
            template_name = entry.attrib["name"]
            for device in entry.xpath("./devices/entry"):
                device_name = device.attrib["name"]
                template_status = next(device.xpath("./template-status/text()"), None)
                status = (template_name, device_name, template_status)
                statuses.append(status)
        return statuses

    def _raw_get_vpn_flows(self, name=None):
        """
        Returns the VPN flow information as a XML object.
        NOTE: This only works on Panorama server, not firewalls
        """
        # This only works on Panorama, not the FW"
        filter = f"<name>{name}</name>" if name else "<all></all>"
        cmd = f"<show><vpn><flow>{filter}</flow></vpn></show>"
        return self.operation(cmd)

    def get_vpn_flows(self, name=None):
        """
        Returns the VPN flow information as a python structure.
        NOTE: This only works on Panorama server, not firewalls
        """
        entries = self._raw_get_vpn_flows(name=name).xpath(".//IPSec/entry")
        return [types.VPNFlow.from_xml(e) for e in entries]

    def system_info(self):
        """
        Returns informations about the system as a XML object.
        """
        cmd = "<show><system><info></info></system></show>"
        return self.operation(cmd)

    def _raw_system_resources(self):
        """
        Get the system resouces as a XML object.
        NOTE: The string is the raw output of a `ps` command.
        """
        cmd = "<show><system><resources></resources></system></show>"
        return self.operation(cmd)

    def system_resources(self):
        """
        Get the system resouces as a string.
        The string is the raw output of a `ps` command.
        """
        res = self._raw_system_resources()
        text = res.xpath(".//result/text()")[0]
        return text.split("\n\n")[0]

    def _raw_download_software(self, version):
        """
        Download the software version on the device.
        version: the software version to download
        """
        cmd = f"<request><system><software><download><version>{version}</version></download></software></system></request>"
        return self.operation(cmd)

    def download_software(self, version) -> Optional[str]:
        """
        Download the software version on the device.
        version: the software version to download

        Returns the download job's ID in case of successful launch,
        None is returned otherwise.
        """
        res = self._raw_download_software(version)
        try:
            return res.xpath(".//job/text()")[0]
        except Exception:
            self.logger.debug("Download has not started")
        return None

    def _raw_install_software(self, version):
        """
        Install the software version on the device.
        version: the software version to install
        """
        cmd = f"<request><system><software><install><version>{version}</version></install></software></system></request>"
        return self.operation(cmd)

    def install_software(
        self, version: Union[None, str, types.SoftwareVersion]
    ) -> Optional[str]:
        """
        Install the software version on the device.
        version: the software version to install

        Returns the download job's ID in case of successful launch,
        None is returned otherwise.
        """
        if isinstance(version, types.SoftwareVersion):
            version = version.version
        res = self._raw_install_software(version)
        try:
            return res.xpath(".//job/text()")[0]
        except Exception:
            self.logger.debug("Download has not started")
        return None

    def _raw_restart(self):
        """
        Restart the device
        """
        cmd = "<request><restart><system></system></restart></request>"
        return self.operation(cmd)

    def restart(self):
        """
        Restart the device
        """
        return "".join(self._raw_restart().xpath(".//result/text()"))

    def automatic_download_software(
        self, version: Optional[str] = None
    ) -> types.SoftwareVersion:
        """
        Automatically download the requested software version.
        if the version is not provided, it defaults to the latest one.

        NOTE: This does not do the installation.
        This is usefull to download in anticipation of the upgrade.
        For automatic install, see `automatic_software_upgrade`
        """
        version_str = version
        versions = self.get_versions()
        sw_version = None
        if not version_str:
            sw_version = next((v for v in versions if v.latest), None)
        else:
            sw_version = next((v for v in versions if v.version == version_str), None)
        if not sw_version:
            self.logger.error(f"Version {version_str} not found")
            return exit(1)

        # Already downloaded: Nothing to do
        if sw_version.downloaded:
            self.logger.info(f"Version {sw_version.version} already downloaded")
            return sw_version

        # Download minor version first (required)
        base_version = next(
            v for v in versions if v.version == sw_version.base_minor_version
        )
        if not base_version.downloaded:
            self.logger.info(
                f"Launching download of minor version {base_version.version}"
            )
            job_id = self.download_software(base_version.version)
            if not job_id:
                raise Exception("Download has not started")
            job = self.wait_job_completion(job_id)
            if job.result != "OK":
                self.logger.debug(job)
                raise Exception(job.details)
            print(job.details)

        # Actually download the wanted version
        self.logger.info(f"Launching download of version {sw_version.version}")
        job_id = self.download_software(sw_version.version)
        if not job_id:
            raise Exception("Download has not started")
        job = self.wait_job_completion(job_id)
        if job.result != "OK":
            self.logger.debug(job)
            raise Exception(job.details)
        self.logger.info(job.details)
        return sw_version

    def automatic_software_upgrade(
        self, version: Optional[str] = None, install: bool = True, restart: bool = True
    ):
        """
        Automatically download and install the requested software version.
        if the version is not provided, it defaults to the latest one.

        NOTE: This does the software install and restart by default.
        If you only want to download, prefer to use `automatic_download_software` method,
        or set install=False. See the parameters for more information.

        install: install the software after the download
        restart: restart the device after the installation. This option is ignored if install=False

        """
        sw_version = self.automatic_download_software(version)
        if sw_version.current:
            self.logger.info(f"Version {sw_version.version} is already installed")
            return sw_version
        if not install:
            return sw_version
        # We may get the following error:
        # "Error: Upgrading from 10.2.4-h10 to 11.1.2 requires a content version of 8761 or greater and found 8638-7689."
        # This should never happen, we decided to report the error and handle this manually
        self.logger.info(f"Launching install of version {sw_version.version}")

        job_id = self.install_software(sw_version.version)
        if not job_id:
            self.logger.error("Install has not started")
            raise Exception("Install has not started")
        job = self.wait_job_completion(job_id)
        self.logger.info(job.details)

        # Do not restart if install failed
        if job.result != "OK":
            self.logger.error("Failed to install software version")
            return sw_version

        if restart:
            self.logger.info("Restarting the device")
            restart_response = self.restart()
            self.logger.info(restart_response)
        return sw_version



from typing import Optional

from pydantic import BaseModel, ConfigDict

from nagra_panorama_api.utils import (
    first,
)
from nagra_panorama_api.xmlapi.types.utils import (
    Datetime,
    mksx,
)
from nagra_panorama_api.xmlapi.utils import (
    el2dict,
)


class Device(BaseModel):
    model_config = ConfigDict(populate_by_name=True, extra="allow")

    serial: str
    connected: bool
    unsupported_version: bool
    wildfire_rt: bool
    deactivated: Optional[str] = None
    hostname: Optional[str] = None
    ip_address: Optional[str] = None
    ipv6_address: Optional[str] = None
    mac_addr: Optional[str] = None
    uptime: Optional[str] = None
    family: Optional[str] = None
    model: Optional[str] = None
    sw_version: Optional[str] = None
    app_version: Optional[str] = None
    av_version: Optional[str] = None
    device_dictionary_version: Optional[str] = ""
    wildfire_version: Optional[str] = None
    threat_version: Optional[str] = None
    url_db: Optional[str] = None
    url_filtering_version: Optional[str] = None
    logdb_version: Optional[str] = None
    vpnclient_package_version: Optional[str] = None
    global_protect_client_package_version: Optional[str] = None
    prev_app_version: Optional[str] = None
    prev_av_version: Optional[str] = None
    prev_threat_version: Optional[str] = None
    prev_wildfire_version: Optional[str] = None
    prev_device_dictionary_version: Optional[str] = ""
    # domain/: str
    # slot_count: str
    # type/: str
    # tag/: str
    # plugin_versions
    # ha_cluster
    ha_peer_serial: Optional[str] = None
    vpn_disable_mode: bool
    operational_mode: str
    certificate_status: Optional[str] = None
    certificate_subject_name: Optional[str] = None
    certificate_expiry: Datetime
    connected_at: Datetime
    custom_certificate_usage: Optional[str] = None
    multi_vsys: bool
    # vsys
    last_masterkey_push_status: str
    last_masterkey_push_timestamp: Optional[str] = None
    express_mode: bool
    device_cert_present: Optional[str] = None
    device_cert_expiry_date: str

    @staticmethod
    def from_xml(xml) -> Optional["Device"]:
        if isinstance(xml, (list, tuple)):
            xml = first(xml)
        if xml is None:
            return None
        res = {k.replace("-", "_"): v for k, v in el2dict(xml)["entry"].items()}
        p = mksx(xml)
        res.update({"ha_peer_serial": p(".//ha/peer/serial/text()")})
        return Device(**res)
        # res = Device(**res)
        # if res.ip_address == "10.9.158.246":
        #     pprint(xml)
        # return res


class VPNFlow(BaseModel):
    model_config = ConfigDict(populate_by_name=True, extra="allow")

    name: str
    id: int
    gwid: int
    inner_if: str
    outer_if: str
    localip: str
    peerip: str
    state: str
    mon: str
    owner: str

    @staticmethod
    def from_xml(xml):
        if isinstance(xml, (list, tuple)):
            xml = first(xml)
        if xml is None:
            return None
        res = {k.replace("-", "_"): v for k, v in el2dict(xml)["entry"].items()}
        return VPNFlow(**res)


class HAInfo(BaseModel):
    model_config = ConfigDict(populate_by_name=True, extra="allow")

    enabled: bool
    preemptive: Optional[bool] = None
    mode: Optional[str] = None
    state: Optional[str] = None
    peer_state: Optional[str] = None
    priority: Optional[int] = None
    peer_priority: Optional[int] = None
    is_primary: Optional[bool] = None
    peer_conn_status: Optional[str] = None
    mgmt_ip: Optional[str] = None
    ha1_ipaddr: Optional[str] = None
    ha1_backup_ipaddr: Optional[str] = None
    ha2_ipaddr: Optional[str] = None
    ha1_macaddress: Optional[str] = None
    ha1_backup_macaddress: Optional[str] = None
    ha2_macaddress: Optional[str] = None

    @staticmethod
    def from_xml(xml):
        if isinstance(xml, (list, tuple)):
            xml = first(xml)
        if xml is None:
            return None
        p = mksx(xml)
        priority = p("./group/local-info/priority/text()", parser=int)
        peer_priority = p("./group/peer-info/priority/text()", parser=int)
        is_primary = None
        if priority is not None and peer_priority is not None:
            is_primary = priority < peer_priority
        return HAInfo(
            enabled=p("./enabled/text()"),
            preemptive=p("./group/local-info/preemptive/text()"),
            mode=p("./group/local-info/mode/text()"),
            state=p("./group/local-info/state/text()"),
            peer_state=p("./group/peer-info/state/text()"),
            priority=priority,
            peer_priority=peer_priority,
            is_primary=is_primary,
            peer_conn_status=p("./group/peer-info/conn-status/text()"),
            mgmt_ip=p("./group/local-info/mgmt-ip/text()"),
            ha1_ipaddr=p("./group/local-info/ha1-ipaddr/text()"),
            ha1_backup_ipaddr=p("./group/local-info/ha1-backup-ipaddr/text()"),
            ha2_ipaddr=p("./group/local-info/ha2-ipaddr/text()"),
            ha1_macaddress=p("./group/local-info/ha1-macaddr/text()"),
            ha1_backup_macaddress=p("./group/local-info/ha1-backup-macaddr/text()"),
            ha2_macaddress=p("./group/local-info/ha2-macaddr/text()"),
        )



from datetime import datetime
from typing import Any, List, Literal, Optional, Union

from pydantic import AliasChoices, BaseModel, ConfigDict, Field, field_validator

from nagra_panorama_api.xmlapi.utils import el2dict as xml2dict


class RuleHit(BaseModel):
    name: str = Field(validation_alias=AliasChoices("@name", "name"))
    device_group: str = Field(alias="device-group")
    rulebase: Literal["pre-rulebase", "post-rulebase"] = Field(alias="rulebase")
    type: str = Field(alias="type")
    state: Optional[str] = Field(alias="rule-state")
    modification: datetime = Field(alias="rule-modification-timestamp")
    creation: datetime = Field(alias="rule-creation-timestamp")
    all_connected: bool = Field(alias="all-connected")

    @field_validator("modification", "creation", mode="before")
    @classmethod
    def ensure_datetime(cls, v: Any):
        if not isinstance(v, int):
            v = int(v)
        return datetime.fromtimestamp(v)

    @staticmethod
    def from_tuple(t):
        device_group, rulebase, rule_type, xml = t
        return RuleHit.model_validate(
            {
                "device-group": device_group,
                "rulebase": rulebase,
                "type": rule_type,
                **xml2dict(xml)["entry"],
            }
        )


class RuleUse(BaseModel):
    model_config = ConfigDict(extra="allow")
    name: str = Field(validation_alias=AliasChoices("@name", "name"))
    description: Optional[str] = Field(default=None)
    uuid: str = Field(validation_alias=AliasChoices("@uuid", "uuid"))
    state: Optional[str] = Field(alias="rule-state")
    bytes: Optional[int] = Field(default=None)
    group_tag: Optional[str] = Field(alias="group-tag", default=None)
    tag: Optional[List[str]] = Field(default=None)
    disabled: Optional[bool] = Field(default=None)
    rule_type: Optional[Literal["interzone", "universal"]] = Field(
        alias="rule-type", default=None
    )
    nat_type: Optional[Literal["ipv4", "ipv6"]] = Field(alias="nat-type", default=None)
    modification: datetime = Field(alias="rule-modification-timestamp")
    creation: datetime = Field(alias="rule-creation-timestamp")

    action: Optional[Union[Literal["allow", "deny", "reset-client"], dict]] = Field(
        default=None
    )
    to_interface: str = Field(alias="to-interface", default=None)
    protocol: Optional[Literal["tcp", "udp"]] = Field(default=None)
    port: Optional[str] = Field(default=None)  # Can be a port range

    to_: Optional[List[str]] = Field(alias="from", default=None)
    from_: Optional[List[str]] = Field(alias="to", default=None)
    source: Optional[List[str]] = Field(default=None)
    destination: Optional[List[str]] = Field(default=None)
    source_translation: Optional[List[str]] = Field(
        alias="source-translation", default=None
    )
    destination_translation: Optional[List[str]] = Field(
        alias="destination-translation", default=None
    )

    source_user: Optional[List[str]] = Field(alias="source-user", default=None)
    application: Optional[List[str]] = Field(default=None)
    category: Optional[List[str]] = Field(default=None)
    service: Optional[List[str]] = Field(default=None)

    icmp_unreachable: Optional[bool] = Field(alias="icmp-unreachable", default=None)
    log_start: Optional[bool] = Field(alias="log-start", default=None)
    log_end: Optional[bool] = Field(alias="log-end", default=None)
    negate_source: Optional[bool] = Field(alias="negate-source", default=None)
    negate_destination: Optional[bool] = Field(alias="negate-destination", default=None)

    @field_validator(
        "tag",
        "to_",
        "from_",
        "source",
        "destination",
        "source_translation",
        "destination_translation",
        "source_user",
        "application",
        "category",
        "service",
        mode="before",
    )
    @classmethod
    def ensure_membership(cls, v: Any):
        if v is None:
            return None
        members = v.get("member") if isinstance(v, dict) else v
        if isinstance(members, str):
            members = [members]
        return members



import enum
import logging
from dataclasses import dataclass
from datetime import datetime, time
from typing import Optional

from nagra_panorama_api.utils import (
    first,
)
from nagra_panorama_api.xmlapi.types.utils import (
    mksx,
    parse_datetime,
    parse_time,
    pd,
)


def parse_tdeq(d):
    if "null" in d:
        return None
    try:
        return parse_time(d)
    except Exception as e:
        logging.debug(e)
    return parse_datetime(d)


def parse_progress(progress):
    try:
        return float(progress)
    except Exception as e:
        logging.debug(f"{e} => Fallback to datetime parsing")

    # When finished, progress becomes the date of the end
    if parse_datetime(progress):
        return 100.0
    return None


class JobResult(enum.Enum):
    OK = "OK"
    FAIL = "FAIL"


@dataclass
class Job:
    tenq: datetime
    tdeq: time
    id: str
    user: str
    type: str
    status: str
    queued: bool
    stoppable: bool
    result: str
    tfin: datetime
    description: str
    position_in_queue: int
    progress: float
    details: str
    warnings: str

    @staticmethod
    def from_xml(xml) -> Optional["Job"]:
        if isinstance(xml, (list, tuple)):
            xml = first(xml)
        if xml is None:
            return None
        p = mksx(xml)
        return Job(
            p("./tenq/text()", parser=pd),
            p("./tdeq/text()", parser=parse_tdeq),
            p("./id/text()"),
            p("./user/text()"),
            p("./type/text()"),
            p("./status/text()"),
            p("./queued/text()") != "NO",
            p("./stoppable/text()") != "NO",
            p("./result/text()"),
            p("./tfin/text()", parser=pd),
            p("./description/text()"),
            p("./positionInQ/text()", parser=int),
            p("./progress/text()", parser=parse_progress),
            "\n".join(xml.xpath("./details/line/text()")),
            p("./warnings/text()"),
        )



import logging
from typing import Union

import requests

from nagra_panorama_api.constants import SUCCESS_CODE

from .utils import Element, detach, etree_fromstring, etree_tostring


def get_tree(
    host, api_key, remove_blank_text=True, verify=False, timeout=None, logger=None
):
    """
    This function returns the whole configuration of Panorama/the firewall
    as an XML object

    remove_blank_text:
        the response might contain extra white-spaces that will mess up
        the display. This option allows to automatically remove the extra white-spaces
    verify: Force the validation of https certificate
    """
    if logger is None:
        logger = logging
    res = requests.get(
        f"{host}/api?type=config&action=show&xpath=/config",
        headers={"X-PAN-KEY": api_key},
        verify=verify,
        timeout=timeout,
    )
    root_tree = etree_fromstring(res.content, remove_blank_text=remove_blank_text)
    try:
        tree = root_tree.xpath("/response/result/config")[0]
    except Exception:
        logger.warning("Response doesn't contains the config tag. Response:")
        logger.warning(etree_tostring(root_tree, pretty_print=True).decode()[:1000])
        raise Exception("Response doesn't contains the config tag. Check the logs")
    detach(tree)  # Remove the /response/result part
    return tree


def parse_msg_result(result: Element) -> str:
    """
    Parse the `msg` tag from Panorama response.
    This function always return the message as a single string
    """
    # Get the content from the default location
    msg = "".join(result.xpath("./msg/text()"))
    # Check another possible location
    if not msg:
        msg = "".join(result.xpath("./result/msg/text()"))
    if msg:
        return msg
    # Last possibility:
    # The message might be split in <line> elements inside of msg
    return "\n".join(result.xpath("./msg/line/text()"))


def _get_rule_use_cmd(device_group, position, rule_type, start_index, number) -> str:
    """
    This function return the `cmd` parameter to filter the data for rule usage.
    NOTE: These values are paged, we need to provide an offset and a limit.

    device_group: the device group that contains the rules and their usage data
    position: when the rule is used. Can be "pre" or "post"
    rule_type: One of the rule-type ("security", "pbf", "nat", "application-override")
    start_index: start index of the records to retrieve (aka "offset")
    number: number of records to retrieve (aka "limit")
    """
    # positions = ("pre", "post")
    return f"""<show><policy-app>
        <mode>get-all</mode>
        <filter>(rule-state eq 'any')</filter>
        <vsysName>{device_group}</vsysName>
        <position>{position}</position>
        <type>{rule_type}</type>
        <anchor>{start_index}</anchor>
        <nrec>{number}</nrec>
        <pageContext>rule_usage</pageContext>
    </policy-app></show>"""


def raw_request(
    url,
    type,
    method="GET",
    vsys=None,
    params=None,
    headers=None,
    remove_blank_text=True,
    verify=False,
    logger=None,
    parse=True,
    stream=None,
    timeout=None,
) -> Union[str, Element]:
    """
    This function is a wrapper around requests.method.
    It returns:
    - XML in case of success
    - a string with the message if a message is expected

    It does the heavy lifting when receiving a response from Panorama:
    - Data clean up
    - Conversion to XML
    - Status check and automaticly raise exception in case of error.
    """
    if logger is None:
        logger = logging
    query_params = {"type": type}
    if params:
        query_params = {**query_params, **params}
    if vsys is not None:
        query_params["vsys"] = vsys

    res = requests.request(
        method=method,
        url=url,
        params=query_params,
        headers=headers,
        verify=verify,
        stream=stream,
        timeout=timeout,
    )
    if not parse:
        return res
    content = res.content.decode()
    tree = etree_fromstring(content, remove_blank_text=remove_blank_text)
    status = tree.attrib["status"]
    code = int(tree.get("code", SUCCESS_CODE))
    msg = parse_msg_result(tree)
    if status == "error" or code < SUCCESS_CODE:
        logger.debug(content[:500])
        raise Exception(msg)
    if msg:
        return msg
    return tree



"""
This example will download and install the latest version of the software.

Python requirements:
- ping3
"""

import logging
from datetime import datetime
from multiprocessing.pool import ThreadPool

import click

# from ping3 import ping  # Ping
import nagra_panorama_api
from nagra_panorama_api.xmlapi.utils import etree_tostring

from .common import ensure_group_load, ensure_output_directory
from .utils import get_logger, getenv, slugify

# set_default_logger()
logging.getLogger().setLevel(logging.INFO)
# logging.getLogger().setLevel(logging.DEBUG)


################################################################################


DEFAULT_PREFIX = "scripted"


def dump_states_worker(args):
    try:
        (
            fw_name,
            fw_host,
            apikey,
            save,
            prefix,
            output,
        ) = args

        logger = get_logger(f"Device {fw_name}")

        now = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        snapshot_name = f"{prefix}_{now}"

        client = nagra_panorama_api.XMLApi(fw_host, apikey, logger=logger)
        if save:
            save_response = client.save_config(snapshot_name)
            logger.info(f"Saved result: {save_response}")
        if not output:
            return True

        fw_output_dir = output / fw_name
        fw_output_dir.mkdir(parents=True, exist_ok=True)

        config_file = fw_output_dir / f"{prefix}_config_{now}.xml"
        if save:
            configuration = etree_tostring(
                client.get_named_configuration(snapshot_name)
            ).decode()
        else:
            configuration = client.export_configuration().decode()
        with config_file.open("w") as f:
            f.write(configuration)
        logger.info(f"Configuration saved here: {config_file}")

        state_file = fw_output_dir / f"{prefix}_state_{now}.tgz"
        state = client.export_device_state()
        with state_file.open("wb") as f:
            f.write(state)
        logger.info(f"State saved here: {config_file}")
        return True

    except Exception as e:
        logger.error(f"""An unexpected error occured\n{e}""")
        # raise e
        logger.debug(f"Worker args: {args}")
        return False


@click.command(
    "dump-states",
    help="Dump the configuration and device-state of a group of devices",
)
@click.argument(
    "file",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    # help="Configuration file containing the upgrade groups"
)
@click.argument("group")  # help="Upgrade group from the configuration to run"
@click.option(
    "--apikey",
    help="Panorama & Firewall api key (must be the same for both). Use one of the following environment variables: PANORAMA_APIKEY, PANO_APIKEY",
)
@click.option("--save/--no-save", default=True)
@click.option("--prefix", help="Prefix for the configuration/state snapshots")
@click.option(
    "--output",
    default="./confdumps",
    help="Destination folder where to put the downloaded files",
)
def dump_states_cmd(file, group, apikey, save, prefix, output):
    if prefix:
        prefix = slugify(prefix)
    if not prefix:
        prefix = DEFAULT_PREFIX
    if len(prefix) > 12:
        logging.error(
            f"Prefix cannot be longer than 12 caracters. Current prefix: '{prefix}'"
        )
    if not apikey:
        apikey = getenv("PANORAMA_APIKEY", "PANO_APIKEY")
    output = ensure_output_directory(output, group)
    if not output and not save:
        logging.warning("No output nor configuration save requested. Nothing to do")
        exit()
    group_data = ensure_group_load(file, group)

    devices = group_data["devices"]

    pool_data = [
        (
            d["hostname"],
            d["ip_address"],
            apikey,
            save,
            prefix,
            output,
        )
        for d in devices
    ]
    failed = 0
    with ThreadPool(len(pool_data)) as pool:
        for i, upgrade_res in enumerate(
            pool.imap_unordered(dump_states_worker, pool_data), 1
        ):
            logging.info(f"Dumped {i}/{len(pool_data)} configurations")
            if not upgrade_res:
                failed += 1
    if failed > 0:
        logging.error(f"There was {failed} failed upgrades")
    else:
        logging.info("All configuration/states were downloaded successfuly")



# Given a list of subnets,
# Find all NAT rules related to an address in the subnet

from dataclasses import dataclass
from ipaddress import IPv4Network, IPv6Network, ip_network
from itertools import chain
from typing import Optional, Union

from pydantic import AliasChoices, AliasPath, BaseModel, Field
from pydantic.functional_validators import field_validator, model_validator
from typing_extensions import Self

from nagra_panorama_api.xmlapi.utils import el2dict

from .utils import List, String

IPNetwork = Union[IPv4Network, IPv6Network]


@dataclass
class NATRule:
    device_group: str
    name: str
    uuid: str
    ttype: str
    to_members: List[str]
    from_members: List[str]
    src_members: List[str]
    dst_members: List[str]
    src_translations: List[str]
    dst_translations: List[str]

    @property
    def members(self):
        return set(
            chain(
                self.to_members,
                self.from_members,
                self.src_members,
                self.dst_members,
                self.src_translations,
                self.dst_translations,
            )
        )


def _parse_trans_addr(xml):
    static = xml.xpath(".//translated-address/text()")  # for static-ip
    dynamic = xml.xpath(".//translated-address/member/text()")  # for dynamic-ip
    dynamic_and_port = xml.xpath(".//interface-address/ip/text()")
    return static, dynamic, dynamic_and_port


def parse_trans_addr(xml):
    static = []
    dynamic = []
    dynamic_and_port = []
    if not xml:
        return static, dynamic, dynamic_and_port
    if not isinstance(xml, (list, tuple)):
        xml = [xml]
    for e in xml:
        a, b, c = _parse_trans_addr(e)
        static.extend(a)
        dynamic.extend(b)
        dynamic_and_port.extend(c)
    return static, dynamic, dynamic_and_port


def get_trans_addr(xml):
    static, dynamic, dynamic_and_port = parse_trans_addr(xml)
    return [*static, *dynamic]


def xml2nat(xml):
    post_pre = xml.getparent().getparent().getparent()
    device_group = post_pre.getparent().attrib["name"]
    name = xml.attrib["name"]
    uuid = xml.attrib["uuid"]
    ttype = post_pre.tag
    to_members = xml.xpath("./to/member/text()")
    from_members = xml.xpath("./from/member/text()")
    src_members = xml.xpath("./source/member/text()")
    dst_members = xml.xpath("./destination/member/text()")
    src_translations = get_trans_addr(xml.xpath("./source-translation"))
    dst_translations = get_trans_addr(xml.xpath("./destination-translation"))
    return NATRule(
        device_group,
        name,
        uuid,
        ttype,
        to_members,
        from_members,
        src_members,
        dst_members,
        src_translations,
        dst_translations,
    )


def get_ip_network(ip_netmask):
    try:
        if ip_netmask:
            return ip_network(ip_netmask, strict=False)
    except Exception:
        return None


# https://docs.pydantic.dev/latest/concepts/alias/#aliaspath-and-aliaschoices
class Address(BaseModel):
    name: str = Field(validation_alias="@name")
    type: Optional[str] = None
    prefix: Optional[str] = None
    ip_netmask: Optional[str] = Field(
        alias="ip-netmask",
        validation_alias=AliasChoices(
            AliasPath("ip-netmask", "#text"),
            "ip-netmask",
        ),
        default=None,
    )
    ip_network: Optional[IPNetwork] = None
    ip_range: Optional[str] = Field(alias="ip-range", default=None)
    fqdn: Optional[String] = None
    tags: List[String] = Field(
        validation_alias=AliasPath("tag", "member"), default=None
    )

    @field_validator("tags", mode="before")
    @classmethod
    def validate_tags(cls, v) -> List[str]:
        if not v:
            return []
        if not isinstance(v, list):
            return [v]
        return v

    @model_validator(mode="after")
    def validate_ip_network(self) -> Self:
        if self.ip_network is None:
            self.ip_network = get_ip_network(self.ip_netmask)
        if not isinstance(self.ip_network, (IPv4Network, IPv6Network)):
            self.ip_network = None
        return self

    @model_validator(mode="after")
    def validate_type(self) -> Self:
        address_type = None
        if self.prefix:
            address_type = "prefix"
        elif self.ip_netmask:
            address_type = "ip-netmask"
        elif self.ip_range:
            address_type = "ip-range"
        elif self.fqdn:
            address_type = "fqdn"
        self.type = address_type
        return self

    @staticmethod
    def from_xml(xml):
        entry = el2dict(xml)["entry"]
        return Address.model_validate(entry)


def get_nat_membership(tree):
    # nat_rules_xml = tree.xpath(".//nat/rules/entry")
    nat_rules_xml = tree.xpath("./devices/entry/device-group//nat/rules/entry")
    nat_rules = [xml2nat(n) for n in nat_rules_xml]

    # Create a dict member -> associated nat rules
    nat_membership = {}
    for n in nat_rules:
        for m in n.members:
            member_nat_rules = nat_membership.setdefault(m, [])
            member_nat_rules.append(n)
    return nat_membership


def find_addresses(tree):
    # addresses_xml = tree.xpath(".//address/entry")
    addresses_xml = tree.xpath("./devices/entry/device-group//address/entry")
    address_objects = [Address.from_xml(n) for n in addresses_xml]

    addresses = []
    subnets = []
    for a in address_objects:
        network = a.ip_network
        # We do not consider ip ranges for now
        if not network:
            continue
        if network.prefixlen == network.max_prefixlen:
            addresses.append(a)
        else:
            subnets.append(a)
    return addresses, subnets



import logging
import typing
from datetime import datetime
from typing import Annotated, Any, TypeVar

from pydantic.functional_validators import (
    BeforeValidator,
    PlainValidator,
)
from typing_extensions import TypeAliasType

from nagra_panorama_api.utils import (
    first,
)

DATETIME_FORMAT = "%Y/%m/%d %H:%M:%S"
TIME_FORMAT = "%H:%M:%S"


def parse_datetime(d):
    try:
        if d is None:
            return d
        return datetime.strptime(d, DATETIME_FORMAT)
    except Exception as e:
        logging.debug(e)
        logging.debug(f"Failed to parse {d} as datetime")
    return d


def parse_time(d):
    return datetime.strptime(d, TIME_FORMAT).time()


# https://docs.pydantic.dev/latest/concepts/types/#custom-types
# JobProgress = TypeAliasType('JobProgress', PlainValidator(parse_progress))
Datetime = TypeAliasType(
    "Datetime", Annotated[datetime, PlainValidator(parse_datetime)]
)


def single_xpath(xml, xpath, parser=None, default=None):
    try:
        res = xml.xpath(xpath)
        res = first(res, None)
    except Exception:
        return default
    if res is None:
        return default
    if not isinstance(res, str):
        res = res.text
    if parser:
        res = parser(res)
    return res


pd = parse_datetime
sx = single_xpath


def mksx(xml):
    def single_xpath(xpath, parser=None, default=None):
        res = sx(xml, xpath, parser=parser, default=default)
        logging.debug(res)
        return res

    return single_xpath


def ensure_list(v: Any) -> typing.List[Any]:
    if v is None:
        return []
    if isinstance(v, list):
        return v
    return [v]


# https://docs.pydantic.dev/latest/concepts/types/#generics
Element = TypeVar("Element", bound=Any)
# Similar to typing.List, but ensure to always return a list
List = Annotated[typing.List[Element], BeforeValidator(ensure_list)]

# from pydantic import TypeAdapter
# ta = TypeAdapter(List[int])
# print(ta.validate_python(5))
# print(ta.validate_python([5]))
# print(ta.validate_python())


def ensure_str(v: Any) -> str:
    if v is None:
        return ""
    if isinstance(v, dict):
        return v["#text"]
    return v


def validate_ip(v: Any) -> str:
    if v is None:
        return ""
    if isinstance(v, dict):
        return v["@name"]
    return v


String = Annotated[str, BeforeValidator(ensure_str)]
Ip = Annotated[str, BeforeValidator(validate_ip)]



import click

from .dump_configs import dump_states_cmd
from .dump_dir import dump_dir
from .software_update import (
    generate_apikey_cmd,
    generate_configuration_file_cmd,
    upgrade_cmd,
)


# Keep it as the main entry point
@click.group()
def cli():
    pass


# Add here your subcommands if needed
cli.add_command(dump_states_cmd)
cli.add_command(upgrade_cmd)
cli.add_command(generate_configuration_file_cmd)
cli.add_command(generate_apikey_cmd)
cli.add_command(dump_dir)



import logging

from nagra_network_misc_utils.logger import set_default_logger

from .commands import cli

set_default_logger()
logging.getLogger().setLevel(logging.INFO)

if __name__ == "__main__":
    cli()



from dataclasses import dataclass
from datetime import datetime

from nagra_panorama_api.utils import (
    first,
)
from nagra_panorama_api.xmlapi.types.utils import (
    mksx,
    pd,
)


@dataclass
class SoftwareVersion:
    version: str
    filename: str
    released_on: datetime
    downloaded: bool
    current: bool
    latest: bool
    uploaded: bool

    @staticmethod
    def from_xml(xml):
        if isinstance(xml, (list, tuple)):
            xml = first(xml)
        if xml is None:
            return None
        p = mksx(xml)
        return SoftwareVersion(
            p("./version/text()"),
            p("./filename/text()"),
            p("./released-on/text()", parser=pd),
            p("./downloaded/text()") != "no",
            p("./current/text()") != "no",
            p("./latest/text()") != "no",
            p("./uploaded/text()") != "no",
        )

    @property
    def base_minor_version(self) -> str:
        major, minor, _ = self.version.split(".")
        return f"{major}.{minor}.0"

    @property
    def base_major_version(self) -> str:
        major, _, _ = self.version.split(".")
        return f"{major}.0.0"



from panos import panorama

from .utils import clean_url_host


class Panorama(panorama.Panorama):
    """
    Wrapper class for the Panorama class from pan-os-python library:
    https://pan-os-python.readthedocs.io/en/latest/readme.html#features

    Added features are:
    - hostname can be provided with and without the scheme
        - https://mydomain.com
        - mydomain.com
        - https://mydomain.com:443
      Are all valid
    """

    def __init__(
        self,
        hostname,
        api_username=None,
        api_password=None,
        api_key=None,
        port=None,
        *args,
        **kwargs,
    ):
        _, hostname, _port = clean_url_host(hostname)
        port = port or _port or 443
        return super().__init__(
            hostname, api_username, api_password, api_key, port, *args, **kwargs
        )



"""
This example will download and install the latest version of the software.

Python requirements:
- ping3
"""

import logging
from multiprocessing.pool import ThreadPool
from pathlib import Path
from typing import List, Optional

import click
import requests
import urllib3
import yaml

# from nagra_network_misc_utils.logger import set_default_logger
# from ping3 import ping  # Ping
import nagra_panorama_api
from nagra_panorama_api.utils import first
from nagra_panorama_api.xmlapi.utils import wait

from .common import DEFAULT_OUTPUT_DIRECTOR, ensure_group_load, ensure_output_directory
from .utils import get_logger, getenv

# set_default_logger()
logging.getLogger().setLevel(logging.INFO)
# logging.getLogger().setLevel(logging.DEBUG)


def wait_availability(client, fw_host, logger=None):
    if logger is None:
        logger = logging
    logger.info(f"Checking firewall '{fw_host}' availability. Waiting for response...")
    for _ in wait():
        try:
            versions = client.get_versions()
            current = next(v for v in versions if v.current)
            if current is None:
                logger.warning("Device is not not answering")
            else:
                return current
        except (urllib3.exceptions.MaxRetryError, requests.exceptions.ConnectionError):
            logger.warning(f"Firewall {fw_host} is still not responding")
        except Exception as e:
            logger.debug(
                f"Unexpected error of type {type(e)} occured on firewall {fw_host}"
            )
            logger.error(f"Firewall {fw_host} is still not responding: {e}")
    raise Exception(f"Timeout while waiting for availability of firewall {fw_host}")


def check_availability(client, fw_host, logger=None):
    if logger is None:
        logger = logging
    ## Wait for the FW to respond
    version = wait_availability(client, fw_host, logger=logger)
    if not version:
        logger.error(f"Device {fw_host} never responded")
        return False
    logger.info(f"Firewall {fw_host} is available")
    return True


def check_panorama_connection(panorama_client, fw_host, logger=None):
    if logger is None:
        logger = logging
    ## Check that the device is connected to Panorama
    devices = panorama_client.get_devices()
    candidates = [d for d in devices if d.ip_address == fw_host]
    if len(candidates) > 1:
        serials = ", ".join(c.serial for c in candidates)
        logger.warning(
            f"Device {fw_host} has {len(candidates)} serial number: {serials}"
        )
    if not candidates:
        logger.error(f"Device with IP {fw_host} is not detected AT ALL to panorama")
        return False
    if not any(c.connected for c in candidates):
        logger.error(
            f"Device with IP {fw_host} is not detected as connected to panorama"
        )
        return False
    logger.info(f"Firewall {fw_host} is detected has connected to panorama")
    return True


def run_checks(
    client,
    fw_host,
    panorama_client: Optional[nagra_panorama_api.XMLApi] = None,
    logger=None,
):
    if logger is None:
        logger = logging
    if not check_availability(client, fw_host, logger=logger):
        return False
    if panorama_client and not check_panorama_connection(
        panorama_client, fw_host, logger=logger
    ):
        return False
    return True


################################################################################
def get_ha_info_worker(args):
    (
        device,
        apikey,
    ) = args
    fw_name, fw_host = device["hostname"], device["ip_address"]
    try:
        client = nagra_panorama_api.XMLApi(fw_host, apikey)
        return device, client.get_ha_info()
    except Exception as e:
        logging.error(
            f"""\
An unexpected error occured on device {fw_name}:
{e}
"""
        )
        return None


def check_and_complete_data(panorama: nagra_panorama_api.XMLApi, devices: List[dict]):
    """
    This function check that all devices are connected to panorama.
    It also add the device information from panorama into the devices object.
    The devices dicts are updated in place and returned at the end of the function
    """
    errors = []
    panorama_devices = panorama.get_devices(connected=True)
    device_map = {d.ip_address: d for d in panorama_devices}
    errors = []
    for d in devices:
        panorama_data = device_map.get(d["fw_host"])
        if panorama_data is None:
            errors.append(d)
            continue
        d["panorama_data"] = panorama_data
    if errors:
        for d in errors:
            logging.error(
                f"Device {d['fw_name']} ({d['fw_host']}) is not connected to panorama"
            )
        exit(1)
    return devices


def check_ha_pairs(panorama, devices):
    """
    NB: This function need to be called after check_and_complete_data

    This function check if we are trying to update a device in an HA pair.
    If yes, it ensures that all devices in the HA pair are set for upgrade.

    """
    errors = []
    ha_pairs, without_ha = panorama.get_ha_pairs(connected=True)
    devices_map = {d["panorama_data"].serial: d for d in devices}
    for pair in ha_pairs:
        # Ensure that each pair contains exactly 2 devices
        if len(pair) != 2:
            members = (
                ", ".join(d.ip_address for d in pair)
                if pair
                else "(NO MEMBER -> Script bug)"
            )
            errors.append(
                f"""\
An HA pair doesn't have 2 active devices. Found members:
{members}
"""
            )
            continue
        # Ensure that we upgrade all devices in an HA pair
        member1, member2 = pair
        member1_found = member1.serial in devices_map
        member2_found = member2.serial in devices_map
        # If any member is found, ensure that member1 is the one found
        # Nb: This implies that
        # - if member1 is not found, none are found
        # - if member2 is found, both are found
        # - if they are not both found, member1 is the one found and member2 is not found
        if member2_found is True:
            member1_found, member2_found = member2_found, member1_found
            member1, member2 = member2, member1
        # If they are not both found/not found, there is an issue
        # Nb: member1 would necessarily be the one found in this situation
        if member1_found != member2_found:
            errors.append(
                f"Device {member1.ip_address} is part of an HA pair "
                f"but its peer {member2.ip_address} is not set for upgrade"
            )
    if errors:
        for e in errors:
            logging.error(e)
        exit(1)

    # Split the devices depending on their HA membership
    # and group the device from the same HA pair

    ha_pairs, without_ha = [], []
    devices_map = {d["panorama_data"].serial: d for d in devices}
    done = set()
    for d in devices:
        data = d["panorama_data"]
        serial = data.serial
        if serial in done:
            continue
        ha_peer_serial = data.ha_peer_serial
        if not ha_peer_serial:
            without_ha.append(d)
            done.add(serial)
            continue
        peer = devices_map.get(ha_peer_serial)
        ha_pairs.append((d, peer))
    return ha_pairs, without_ha


def upgrade_worker(device: dict):
    try:
        logger = device["logger"]
        panorama_host = device["panorama_host"]
        fw_name = device["fw_name"]
        fw_host = device["fw_host"]
        apikey = device["apikey"]
        version = device["version"]
        download_only = device["download_only"]
        test_only = device["test_only"]

        panorama_client = None
        if panorama_host:
            panorama_client = nagra_panorama_api.XMLApi(
                panorama_host, apikey, logger=logger
            )
        client = nagra_panorama_api.XMLApi(fw_host, apikey, logger=logger)

        fw_sys_info = client.system_info()
        serial_number = fw_sys_info.xpath(".//serial/text()")[0]
        current_version = fw_sys_info.xpath(".//sw-version/text()")[0]
        logger.info(
            f"Device with IP {fw_host} (S/N: {serial_number}) is currently on version '{current_version}'"
        )
        if not test_only:
            version = client.automatic_software_upgrade(
                version, install=(not download_only)
            )

        # Tests
        res = run_checks(
            client, fw_host, panorama_client=panorama_client, logger=logger
        )
        if download_only:
            logger.info(
                f"Done dowloading version { version.version } on device {fw_name} ({fw_host}) "
            )
        elif not test_only:
            logger.info(
                f"Done upgrading device {fw_name} ({fw_host}) to version { version.version }"
            )
        return res
    except Exception as e:
        logger.error(
            f"""\
An unexpected error occured
Worker args: {device}
{e}
"""
        )
        return False


def group_by_panorama(devices: List[dict]):
    def worker(device):
        fw_name = device["fw_name"]
        fw_host = device["fw_host"]
        apikey = device["apikey"]
        try:
            client = nagra_panorama_api.XMLApi(fw_host, apikey)
            panorama_ip = client.get_local_panorama_ip()
            return True, panorama_ip, device
        except Exception as e:
            logging.error(f"Couldn't find panorama server for device '{fw_name}':\n{e}")
        return True, None, device

    grouped = {}
    failed = []
    with ThreadPool(len(devices)) as pool:
        for success, panorama_ip, device in pool.imap_unordered(worker, devices):
            if not success:
                failed.append(device)
                continue
            device.setdefault("panorama_host", panorama_ip)
            g = grouped.setdefault(panorama_ip, [])
            g.append(device)
    if failed:
        failed_devices = ", ".join(
            sorted(f"{d['fw_name']} ({d['fw_host']})" for d in failed)
        )
        logging.error(
            f"Couldn't assess the Panorama server of the following devices: {failed_devices}"
        )
        # exit(1)  # Stop the upgrade ? Ask for continuation ?
    return grouped


def trim_text(text, max_len=100):
    line_end = "..."
    trim_len = max(max_len - len(line_end), 0)
    if len(text) > max_len:
        text = text[:trim_len] + line_end
    return text


@click.command(
    "upgrade",
    help="Upgrade a group of devices to a different softare version",
)
@click.argument(
    "file",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    # help="Configuration file containing the upgrade groups"
)
@click.argument("group")  # help="Upgrade group from the configuration to run"
@click.option(
    "--apikey",
    help="Panorama & Firewall api key (must be the same for both). Use one of the following environment variables: PANORAMA_APIKEY, PANO_APIKEY",
)
@click.option(
    "--download-only",
    is_flag=True,
    default=False,
    help="Only download the software version without installing it",
)
@click.option("--test-only", is_flag=True, default=False, help="Only runs the checks")
@click.option(
    "--logs-output",
    help=f"Destination folder where to put the logs (Default to {DEFAULT_OUTPUT_DIRECTOR})",
)
def upgrade_cmd(file, group, apikey, download_only, test_only, logs_output):
    if download_only and test_only:
        logging.error(
            "You cannot not use --download-only and --test-only flags at the same time"
        )
        exit(1)
    if not apikey:
        apikey = getenv("PANORAMA_APIKEY", "PANO_APIKEY")

    logs_output = ensure_output_directory(logs_output, group)
    group_data = ensure_group_load(file, group)

    if logs_output:
        generic_logs_file = logs_output / "main.log"
        generic_logs_file_handler = logging.FileHandler(generic_logs_file)
        # generic_logs_file_handler.setLevel(logging.DEBUG)  # Set the logging level for the file handler
        # Create a formatter and set it for the handler
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        generic_logs_file_handler.setFormatter(formatter)
        # Add the handler to the logger
        logging.getLogger().addHandler(generic_logs_file_handler)

    # panorama_ip = group_data["panorama_ip"]
    if "panorama_ip" in group_data:
        logging.warning("Entry 'panorama_ip' is deprecated and will be ignored")

    version = group_data.get("version")
    devices = group_data["devices"]

    # Transform the device lists:
    # - Simplify maintenance (e.g. renaming) and debuging
    # - make each device data self-contained
    devices = [
        {
            "group": group,
            "panorama_host": None,
            "panorama_data": None,  # See check_and_complete_data
            "fw_name": d["hostname"],
            "fw_host": d["ip_address"],
            "apikey": apikey,
            "version": version,
            "download_only": download_only,
            "test_only": test_only,
            "logs_output": logs_output,
        }
        for d in devices
    ]

    # Find panorama server associated to the devices (if there is one)
    panorama_ip = None
    grouped = group_by_panorama(devices)
    if len(grouped) > 1:
        statistics = "\n".join(
            f"""{panorama_server or '<no-server>'} => {len(devices)} ({trim_text(', '.join(d['fw_name'] for d in devices))})"""
            for panorama_server, devices in grouped.items()
        )
        logging.error(
            f"Devices in group {group} are associated to different panorama server:\n{statistics}"
        )

        exit(1)
    elif len(grouped) == 1:
        panorama_ip, devices = first(grouped.items())

    panorama = None
    if panorama_ip:
        panorama = nagra_panorama_api.XMLApi(panorama_ip, apikey)
    else:
        # TODO: Check the HA pair directly from the devices, not from panorama
        logging.warning(
            "No panorama host defined in configuration. Some features won't apply, like checking HA pairs"
        )
    if panorama:
        devices = check_and_complete_data(panorama, devices)

    for d in devices:
        fw_name = d["fw_name"]
        logfile = logs_output / fw_name if logs_output else None
        logger = get_logger(f"Device {fw_name}", logfile)
        d["logger"] = logger

    if panorama:
        ha_pairs, without_ha = check_ha_pairs(panorama, devices)

        devices = without_ha
        if ha_pairs and not (download_only or test_only):
            ha_pairs_textual_list = "\n".join(
                " - ".join(f"{d['fw_name']} ({d['fw_host']})" for d in pair)
                for pair in ha_pairs
            )
            logging.warning(
                f"Group {group} contains devices that are part of HA pairs."
                "The script cannot upgrade them at the moment and will ignore them.\n"
                "You can retry with flags --download-only or --test-only.\n"
                f"List of HA pairs found:\n{ha_pairs_textual_list}"
            )

    if not test_only:
        if not version:
            logging.warning(
                "version not defined, the script will default to the latest version for each devices"
            )
        else:
            logging.info(f"Starting upgrade to version {version}")

    if not devices:
        logging.info("No device to manage.")
        exit()

    operation, op_adj = "upgrades", "Upgraded"
    if test_only:
        operation, op_adj = "tests", "Tested"
    elif download_only:
        operation, op_adj = "downloads", "Downloaded"

    failed_upgrades = 0
    with ThreadPool(len(devices)) as pool:
        for i, upgrade_res in enumerate(
            pool.imap_unordered(upgrade_worker, devices), 1
        ):
            logging.info(f"{op_adj} {i}/{len(devices)}")
            if not upgrade_res:
                failed_upgrades += 1

    if failed_upgrades > 0:
        logging.error(f"There was {failed_upgrades} failed {operation}")
    else:
        logging.info(f"All {operation} went successfuly")


@click.command(
    "generate-apikey",
    help="This command generate the configuration file boilerplate using all devices connected to panorama",
)
@click.option(
    "--panorama-host",
    "host",
    envvar="PANORAMA_HOST",
    required=True,
    help="Panorama host/ip. Use PANORAMA_HOST environment variable",
)
@click.option(
    "--username",
    envvar="PANORAMA_USERNAME",
    required=True,
    help="Username password to access Panorama (envvar: PANORAMA_USERNAME)",
)
@click.option(
    "--password",
    envvar="PANORAMA_PASSWORD",
    prompt=True,
    hide_input=True,
    required=True,
    help="User password to access Panorama (envvar: PANORAMA_PASSWORD)",
)
def generate_apikey_cmd(host, username, password):
    client = nagra_panorama_api.XMLApi(host, "dummykey")
    try:
        key = client.generate_apikey(username, password)
        print(key)
    except Exception as e:
        print(e)
        exit(1)


@click.command(
    "generate-config",
    help="This command generate the configuration file boilerplate using all devices connected to panorama",
)
@click.option(
    "--panorama-host",
    "host",
    help="Panorama host/ip. Use PANORAMA_HOST environment variable",
)
@click.option(
    "--apikey",
    help="Panorama api key. Use one of the following environment variables: PANORAMA_APIKEY, PANO_APIKEY",
)
@click.option(
    "--sw-version",
    "version",
)
@click.option(
    "-o",
    "--out",
    help="Output file for the configuration containing the upgrade groups (default to stdin)",
)
def generate_configuration_file_cmd(host, apikey, version, out):
    if not host:
        host = getenv("PANORAMA_HOST")
    if not apikey:
        apikey = getenv("PANORAMA_APIKEY", "PANO_APIKEY")
    if not apikey:
        logging.error("Missing Panorama API Key")
        exit(1)
    panorama_client = nagra_panorama_api.XMLApi(host, apikey)
    connected_devices = panorama_client.get_devices(connected=True)
    data = {
        "groups": {
            "all": {
                # "panorama_ip": host,
                "version": version,
                "devices": [
                    {
                        "hostname": d.hostname,
                        "ip_address": d.ip_address,
                    }
                    for d in connected_devices
                ],
            }
        },
    }

    content = yaml.dump(data, sort_keys=False, explicit_start=True, indent=2)
    if not out:
        print(content)
    else:
        output = Path(out)
        output.write_text(content)


# client = nagra_panorama_api.XMLApi(os.environ["PANO_HOST"], os.environ["PANO_APIKEY"])
# fw_sys_info = client.system_info()

# vpn_flows = client.get_vpn_flows()
# template_status = list(panorama_client.get_templates_sync_status())



# Given a list of subnets,
# Find all NAT rules related to an address in the subnet

from ipaddress import ip_network
from typing import ClassVar, Iterable, Optional

from pydantic import AliasPath, BaseModel, ConfigDict, Field

from nagra_panorama_api.xmlapi.utils import el2dict

from .utils import Ip, List, String


def get_ip_network(ip_netmask):
    try:
        if ip_netmask:
            return ip_network(ip_netmask, strict=False)
    except Exception:
        return None


class GenericInterface(BaseModel):
    model_config = ConfigDict(extra="ignore")
    inttype: str
    parent: Optional[str] = None
    name: str
    description: String = ""
    comment: String = ""
    ip: List[str] = Field(default_factory=list)
    untagged_sub_interface: Optional[String] = None
    tags: List[String] = Field(default_factory=list)
    link_state: Optional[str] = None
    link_speed: Optional[str] = None
    link_duplex: Optional[str] = None
    aggregate_group: Optional[String] = None


class Layer2(BaseModel):
    model_config = ConfigDict(extra="allow")
    inttype: ClassVar[str] = "layer2"

    name: String = Field(validation_alias="@name", default="")
    units: List["Layer2"] = Field(
        alias="units",
        validation_alias=AliasPath("units", "entry"),
        default_factory=list,
    )
    comment: String = ""
    tags: List[String] = Field(
        validation_alias=AliasPath("tag", "member"),
        default_factory=list,
    )

    def flatten(self, parent=None) -> Iterable[GenericInterface]:
        if self.name:
            dump = self.model_dump()
            dump.update(
                {
                    "inttype": self.inttype,
                    "parent": parent,
                }
            )
            generic = GenericInterface.model_validate(dump)
            yield generic
            parent = self.name
        for u in self.units:
            yield from u.flatten(parent)


class Layer3(BaseModel):
    model_config = ConfigDict(extra="allow")
    inttype: ClassVar[str] = "layer3"

    name: Optional[String] = Field(validation_alias="@name", default=None)
    description: String = Field(validation_alias="#text", default="")
    ip: List[Ip] = Field(
        validation_alias=AliasPath("ip", "entry"), default_factory=list
    )
    untagged_sub_interface: Optional[String] = Field(
        alias="untagged-sub-interface",
        default=None,
    )
    units: List["Layer3"] = Field(
        alias="units",
        validation_alias=AliasPath("units", "entry"),
        default_factory=list,
    )
    comment: String = ""
    tags: List[String] = Field(
        validation_alias=AliasPath("tag", "member"),
        default_factory=list,
    )

    def flatten(self, parent=None) -> Iterable[GenericInterface]:
        if self.name:
            dump = self.model_dump()
            dump.update(
                {
                    "inttype": self.inttype,
                    "parent": parent,
                }
            )
            generic = GenericInterface.model_validate(dump)
            yield generic
            parent = self.name
        for u in self.units:
            yield from u.flatten(parent)


class Vlan(BaseModel):
    model_config = ConfigDict(extra="allow")
    inttype: ClassVar[str] = "vlan"

    name: Optional[String] = Field(validation_alias="@name", default=None)
    description: String = Field(validation_alias="#text", default="")
    ip: List[Ip] = Field(
        validation_alias=AliasPath("ip", "entry"), default_factory=list
    )
    # untagged_sub_interface: Optional[String] = Field(
    #     alias="untagged-sub-interface",
    #     default=None,
    # )
    units: List["Vlan"] = Field(
        alias="units",
        validation_alias=AliasPath("units", "entry"),
        default_factory=list,
    )
    comment: String = ""
    tags: List[String] = Field(
        validation_alias=AliasPath("tag", "member"),
        default_factory=list,
    )

    def flatten(self, parent=None) -> Iterable[GenericInterface]:
        if self.name:
            dump = self.model_dump()
            dump.update(
                {
                    "inttype": self.inttype,
                    "parent": parent,
                }
            )
            generic = GenericInterface.model_validate(dump)
            yield generic
            parent = self.name
        for u in self.units:
            yield from u.flatten(parent)


class Ethernet(BaseModel):
    model_config = ConfigDict(extra="allow")
    inttype: ClassVar[str] = "ethernet"

    name: str = Field(validation_alias="@name")
    ip: List[str] = Field(
        validation_alias=AliasPath("layer3", "ip", "entry", "@name"),
        default_factory=list,
    )
    description: String = Field(validation_alias="#text", default="")
    link_state: Optional[str] = Field(
        alias="link-state",
        validation_alias=AliasPath("link-state", "#text"),
        default=None,
    )
    link_speed: Optional[str] = Field(
        alias="link-speed",
        validation_alias=AliasPath("link-speed", "#text"),
        default=None,
    )
    link_duplex: Optional[str] = Field(
        alias="link-duplex",
        validation_alias=AliasPath("link-duplex", "#text"),
        default=None,
    )
    aggregate_group: Optional[String] = Field(
        alias="aggregate-group",
        default=None,
    )
    layer2: Optional[Layer2] = Field(alias="layer2", default=None)
    layer3: Optional[Layer3] = Field(alias="layer3", default=None)
    tags: List[String] = Field(
        validation_alias=AliasPath("tag", "member"),
        default_factory=list,
    )

    def flatten(self, parent=None) -> Iterable[GenericInterface]:
        if self.name:
            dump = self.model_dump()
            dump.update(
                {
                    "inttype": self.inttype,
                    "parent": parent,
                }
            )
            generic = GenericInterface.model_validate(dump)
            yield generic
            parent = self.name
        if self.layer2:
            yield from self.layer2.flatten(parent)
        if self.layer3:
            yield from self.layer3.flatten(parent)

    # @model_validator(mode="before")
    # @classmethod
    # def test(cls, data):
    #     print("\n" * 5)
    #     print("Ethernet:", data)
    #     return data


class AggregateEthernet(BaseModel):
    model_config = ConfigDict(extra="allow")
    inttype: ClassVar[str] = "aggregate"

    name: str = Field(validation_alias="@name")
    ip: List[str] = Field(
        validation_alias=AliasPath("layer3", "ip", "entry", "@name"),
        default_factory=list,
    )
    description: String = Field(validation_alias="#text", default="")
    comment: String = ""
    units: List["AggregateEthernet"] = Field(
        alias="units",
        validation_alias=AliasPath("units", "entry"),
        default_factory=list,
    )
    layer2: Optional[Layer2] = Field(alias="layer2", default=None)
    layer3: Optional[Layer3] = Field(alias="layer3", default=None)
    untagged_sub_interface: Optional[String] = Field(
        alias="untagged-sub-interface",
        default=None,
    )
    tags: List[String] = Field(
        validation_alias=AliasPath("tag", "member"),
        default_factory=list,
    )
    # @model_validator(mode="before")
    # @classmethod
    # def test(cls, data):
    #     print("\n" * 5)
    #     print("AggregateEthernet:", data)
    #     return data

    def flatten(self, parent=None) -> Iterable[GenericInterface]:
        if self.name:
            dump = self.model_dump()
            dump.update(
                {
                    "inttype": self.inttype,
                    "parent": parent,
                }
            )
            generic = GenericInterface.model_validate(dump)
            yield generic
            parent = self.name
        if self.layer2:
            yield from self.layer2.flatten(parent)
        if self.layer3:
            yield from self.layer3.flatten(parent)


class Loopback(BaseModel):
    model_config = ConfigDict(extra="allow")
    inttype: ClassVar[str] = "loopback"

    name: str = Field(validation_alias="@name")
    description: Optional[String] = Field(validation_alias="#text", default=None)
    ip: List[Ip] = Field(
        validation_alias=AliasPath("ip", "entry"), default_factory=list
    )
    comment: Optional[str] = Field(
        alias="comment", validation_alias=AliasPath("comment", "#text"), default=None
    )
    tags: List[String] = Field(
        validation_alias=AliasPath("tag", "member"),
        default_factory=list,
    )

    def flatten(self, parent=None) -> Iterable[GenericInterface]:
        dump = self.model_dump()
        dump.update(
            {
                "inttype": self.inttype,
                "parent": parent,
            }
        )
        generic = GenericInterface.model_validate(dump)
        yield generic


# https://docs.pydantic.dev/latest/concepts/alias/#aliaspath-and-aliaschoices
class Interface(BaseModel):
    model_config = ConfigDict(extra="allow")

    aggregate_ethernet: List[AggregateEthernet] = Field(
        alias="aggregate-ethernet",
        validation_alias=AliasPath("aggregate-ethernet", "entry"),
        default_factory=list,
    )
    # entry = Field(alias="entry")
    ethernet: List[Ethernet] = Field(
        alias="ethernet",
        validation_alias=AliasPath("ethernet", "entry"),
        default_factory=list,
    )
    loopback: List[Loopback] = Field(
        alias="loopback",
        validation_alias=AliasPath("loopback", "units", "entry"),
        default_factory=list,
    )
    vlan: List[Vlan] = Field(
        alias="vlan",
        validation_alias=AliasPath("vlan", "units", "entry"),
        default_factory=list,
    )
    # ha1 = Field(alias='ha1')
    # ha1_backup = Field(alias='ha1-backup')
    # ha2 = Field(alias='ha2')
    # ha2_backup = Field(alias='ha2-backup')
    # ha3 = Field(alias='ha3')
    # member = Field(alias='member')
    # tunnel = Field(alias='tunnel')

    @staticmethod
    def from_xml(xml):
        entry = el2dict(xml)["interface"]
        return Interface.model_validate(entry)

    def _flatten(self) -> Iterable[GenericInterface]:
        for eth in self.ethernet:
            yield from eth.flatten()
        for agg in self.aggregate_ethernet:
            yield from agg.flatten()
        for v in self.vlan:
            yield from v.flatten()
        for lb in self.loopback:
            yield from lb.flatten()

    def flatten(self) -> List[GenericInterface]:
        return list(self._flatten())
